/// Josh Unit Test Library

#pragma once
#include <functional>
#include <string>
#include <vector>
#include <jlog/jlog.hpp>

namespace jtest {

    // Requirements
    //

    // Can't we just store a struct in a global vector per test?
    // -maxine
    struct testdef
    {
        std::string testname;
        std::function<void()> callback;
        std::string file; // <- & is not needed here -maxine
        int line;
        bool passed;
    };

    // Globals for test tracking
    std::vector<testdef> testlist;
    int rantests;
    int passedtests;
    int failedtests;

    void definetest(const std::string& testname, const std::function<void()>& callback, const std::string& file, int line)
    {
        testlist.push_back(testdef(testname, callback, file, line));
    }

    // TODO: Implement check variants
    // TODO: implement streaming a custom failure message with << operator
    // i.e. : check(my_cond) << "The condition is not true!"
    bool check(bool condition) {
        if (!condition)
            throw std::runtime_error("Test check failed!!");
        return condition;
    }

    bool test(const std::string& testname, const std::function<void()>& callback, const std::string& file, int line)
    {
        try {
            callback();
        } catch(const std::exception& e) {
            rantests++;
            failedtests++;
            jlog::log({
                {.colorCode = jlog::ansi_escape_codes::FG_WHITE, .content ="JTEST"},
                {.content = std::format("{}:{}", file, line)},
                {.colorCode = jlog::ansi_escape_codes::FG_RED, .content = testname},
                {.content = "Failed:", .delimiter = ""},
                {.content = std::format("{}/{}", rantests, testlist.size())},
            });
            return false;
        }

        rantests++;
        passedtests++;
        jlog::log({
                {.colorCode = jlog::ansi_escape_codes::FG_WHITE, .content ="JTEST"},
                {.content = std::format("{}:{}", file, line)},
                {.colorCode = jlog::ansi_escape_codes::FG_GREEN, .content = testname},
                {.content = "Passed:", .delimiter = ""},
                {.content = std::format("{}/{}", rantests, testlist.size())},
            });
        return true;
    }

    // Storing a global vector with all the tests should allow us to loop through all the tests
    // We can also possibly do more tracing and allow other fancy features.
    // -maxine
    void run_tests() {
        //int i;
        //for (int i = 1; const testdef& td : testlist)
        for (testdef& td : testlist)
        {
            td.passed = test(td.testname, td.callback, td.file, td.line);
            //i++;
        }

        jlog::log({
                //{.content = std::format("{}:{}", file, line)},
                {.colorCode = jlog::ansi_escape_codes::FG_WHITE, .content = "JTEST"},
                {.content = std::format("Tests Ran: [{}/{}] Failed: [{}/{}] Passed: [{}/{}]", rantests, testlist.size(), failedtests, rantests, passedtests, rantests), .delimiter = ""},
            });
        //USINFO(std::format("Tests Ran: [{}/{}] Failed: [{}/{}] Passed: [{}/{}]", rantests, testlist.size(), failedtests, rantests, passedtests, rantests))
        if (passedtests == rantests)
        {
            USINFO("All tests passed congratulations! Do you wanna cookie?");
            jlog::log({
                //{.content = std::format("{}:{}", file, line)},
                {.colorCode = jlog::ansi_escape_codes::FG_WHITE, .content = "JTEST"},
                {.colorCode = jlog::ansi_escape_codes::FG_GREEN, .content = "All tests passed congratulations! Do you wanna cookie?", .delimiter = ""},
            });
        }
    }

}

//#define TEST(a, b) jtest::test(a, b, __FILE__, __LINE__);
// Same definition as before essentially, but points to a different function which adds the test to the global vector.
// -maxine
#define TEST(a, b) jtest::definetest(a, b, __FILE__, __LINE__);
